#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <ncurses.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <time.h>
#include <syslog.h>
#include <string.h>

#include <sndfile.h>

#include "error.h"
#include "snd_dev.h"
#include "utils.h"

#define SAMPLE_RATE 44100

char *wav_path = "listen_wav_out";			/* path for wav-files */
char *devname = "/dev/dsp";				/* path to audio-device */
char *configfile = "/usr/local/etc/listener.conf";	/* configurationfile */
int detect_level = 754;					/* threshold for sound-detection */
int rec_silence = 2;					/* number of seconds to keep on reading when silence is detected */
int min_duration = 1;					/* how many samples above threshold before recording starts */
int sample_rate = SAMPLE_RATE;				/* speaks for itself */

void start_recording(int fd, short *first, int first_n_samples)
{
	short *buffer;
	SNDFILE *sndfile_out;
	SF_INFO sfinfo_out;
	char fname[PATH_MAX];
	struct tm *tmdummy;
	time_t recstart, prev, now;

	buffer = (short *)mymalloc(sample_rate * sizeof(short), "recording buffer(2)");

	/* create filename */
	time(&recstart);
	tmdummy = localtime(&recstart);
	sprintf(fname, "%s/%04d-%02d-%02d_%02d%02d%02d.wav",
		wav_path,
		tmdummy -> tm_year + 1900,
		tmdummy -> tm_mon + 1,
		tmdummy -> tm_mday,
		tmdummy -> tm_hour,
		tmdummy -> tm_min,
		tmdummy -> tm_sec);
	prev = now = recstart;

	/* create file */
	memset(&sfinfo_out, 0x00, sizeof(sfinfo_out));
        sfinfo_out.samplerate = sample_rate;
        sfinfo_out.channels = 1;
        sfinfo_out.format = SF_FORMAT_WAV | SF_FORMAT_PCM_16;
        sndfile_out = sf_open(fname, SFM_WRITE, &sfinfo_out);
	if (!sndfile_out)
		error_exit("Could not create file %s!", fname);

	syslog(LOG_INFO, "Started recording to %s", fname);

	/* write sample to disk */
	if (sf_writef_short(sndfile_out, first, first_n_samples) != first_n_samples)
		error_exit("failed to write to wav-file\n");

	for(; (now - prev) < rec_silence; )
	{
		int loop, threshold_count = 0;

		/* read sample */
		if (READ(fd, (char *)buffer, sample_rate * 2 * 1) != (sample_rate * 2 * 1))
			error_exit("failed to read from audio-device\n");

		/* write sample to disk */
		if (sf_writef_short(sndfile_out, buffer, sample_rate) != sample_rate)
			error_exit("failed to write to wav-file\n");

		/* see if it became silent again */
		for(loop=0; loop<sample_rate; loop++)
		{
			if (abs(buffer[loop]) > detect_level)
			{
				threshold_count++;

				if (threshold_count >= min_duration)
				{
					time(&prev);
					break;
				}
			}
		}

		time(&now);
	}

	sf_close(sndfile_out);

	free(buffer);

	syslog(LOG_INFO, "Stopped recording to %s (%d seconds)", fname, (int)(now - recstart));
}

int main(int argc, char *argv[])
{
	int fd;
	FILE *fh;
	/* right now I'm not sure if a short is always 2 bytes. check this on your
	 * platform
	 */
	short *buffer;
	int buffer_size;

	printf("listener v" VERSION ", (C)2003-2004 by folkert@vanheusden.com\n");

	/* determine device to open */
	if (argc == 2)
		configfile = argv[1];
	else if (argc > 2)
		error_exit("Usage: %s configfile\n", argv[0]);

	/* load configfile */
	fh = fopen(configfile, "rb");
	if (!fh)
		error_exit("error opening configfile %s\n", configfile);
	for(;;)
	{
		char *i_is, *cmd, *par = NULL;
		char *line = read_line(fh);
		if (!line)
			break;

                // find '='
                i_is = strchr(line, '=');
                if (i_is)
                {
                        par = get_token(i_is+1);
                        *i_is = 0x00;
                }
                cmd = get_token(line);

		if (strcmp(cmd, "wav_path") == 0)
			wav_path = strdup(par);
		else if (strcmp(cmd, "devname") == 0)
			devname = strdup(par);
		else if (strcmp(cmd, "detect_level") == 0)
			detect_level = atoi(par);
		else if (strcmp(cmd, "rec_silence") == 0)
			rec_silence = atoi(par);
		else if (strcmp(cmd, "min_duration") == 0)
			min_duration = atoi(par);
		else if (strcmp(cmd, "sample_rate") == 0)
			sample_rate = atoi(par);
		else
			error_exit("'%s=%s' is not a known configuration-statement\n", cmd?cmd:"(none)", par?par:"(none)");

		free(par);
		free(cmd);
		free(line);
	}
	fclose(fh);

	if (min_duration < 1)
		error_exit("min_duration must be at least 1!");

	/* open audio-device */
	fd = open(devname, O_RDWR);
	if (fd == -1)
		error_exit("Error opening device %s\n", devname);

	/* set audio-device to 44.1KHz 16bit mono */
	set_soundcard_sample_format(fd);
	set_soundcard_number_of_channels(fd, 1);
	set_soundcard_sample_rate(fd, sample_rate);

	if (daemon(-1, -1) == -1)
		error_exit("problem becoming daemon-process\n");

	syslog(LOG_INFO, "listener started");

	buffer_size = sample_rate * sizeof(short);
	buffer = (short *)mymalloc(buffer_size, "recording buffer(1)");

	for(;;)
	{
		int threshold_count = 0;
		int rc, loop;

		/* check audio-level */
		/* read data */
		rc = READ(fd, (char *)buffer, buffer_size);
		if (rc < 0)
			error_exit("error while reading from audio-device!\n", rc);

		/* check audio-levels */
		for(loop=0; loop<sample_rate; loop++)
		{
			if (abs(buffer[loop]) > detect_level)
			{
				threshold_count++;

				if (threshold_count >= min_duration)
				{
					start_recording(fd, buffer, sample_rate);
					break;
				}
			}
		}
	}

	return 0;
}
