#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <unistd.h>
#include <math.h>
#include <ncurses.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "error.h"
#include "snd_dev.h"
#include "utils.h"

#define MAX_CHECK_LEVEL 4096	/* 1/8th of the max. amplitude */
#define SAMPLE_RATE 8000
#define REFRESH_RATE 5
#define SAMPLE_LENGTH (SAMPLE_RATE / REFRESH_RATE)

void init_curses(void)
{
	initscr();
	keypad(stdscr, TRUE);
	cbreak();
	intrflush(stdscr, FALSE);
	leaveok(stdscr, FALSE);
	noecho();
	nonl();
	refresh();
	nodelay(stdscr, FALSE);
	meta(stdscr, TRUE);	/* enable 8-bit input */
#if 0
	raw();	/* to be able to catch ctrl+c */
#endif
}

int main(int argc, char *argv[])
{
	WINDOW *win;
	int fd;
	char *devname = "/dev/dsp";
	int level = 0, prevlevel = -1;
	int detect_level = 0;

	/* determine device to open */
	if (argc == 2)
		devname = argv[1];
	else if (argc > 2)
		error_exit("Usage: %s device\n", argv[0]);

	/* start curses library */
	init_curses();
	if (COLS < 80 || LINES < 24)
		error_exit("Min. terminal size is 80x24!\n");

	start_color();

	/* create window with a border */
	win = subwin(stdscr, 24, 80, 0, 0);
	if (!win)
		error_exit("Failed to create window!\n");

	if (wborder(win, 0, 0, 0, 0, 0, 0, 0, 0) == ERR)
		error_exit("wborder failed\n");

	/* open audio-device */
	fd = open(devname, O_RDWR);
	if (fd == -1)
		error_exit("Error opening device %s\n", devname);

	/* display some info */
	wattron(win, A_REVERSE);
	mvwprintw(win, 2, 2, "Setup listener");
	wattroff(win, A_REVERSE);
	mvwprintw(win, 4, 2, "With this program you can define at what soundlevel");
	mvwprintw(win, 5, 2, "the program listener should start recording.");

	wattron(win, A_REVERSE);
	mvwprintw(win, 8, 1, " ");
	mvwprintw(win, 8, 78, " ");
	wattroff(win, A_REVERSE);

	mvwprintw(win, 13, 2, "What                    | Configurationfile parameter");
	mvwprintw(win, 14, 2, "------------------------+----------------------------");
	mvwprintw(win, 15, 2, "Current detection level | detect_level");
	mvwprintw(win, 16, 2, "Number of triggers      | min_duration");

	init_pair(1, COLOR_YELLOW, COLOR_BLACK);
	wattron(win, COLOR_PAIR(1));
	mvwprintw(win, 6, 2, "Current dection level:");
	mvwprintw(win, 11, 2, "Number of triggers:");
	wattroff(win, COLOR_PAIR(1));
	mvwprintw(win, 9, 2, "Current average level:");

	mvwprintw(win, 22, 42, "(C) 2003-2004 folkert@vanheusden.com");

	/* set audio-device to 8KHz 16bit mono */
	set_soundcard_sample_format(fd);
	set_soundcard_number_of_channels(fd, 1);
	set_soundcard_sample_rate(fd, SAMPLE_RATE);

	for(;;)
	{
		int c;
		struct timeval tv;
		fd_set rfds;
		/* right now I'm not sure if a short is always 2 bytes. check this on your
		 * platform
		 */
		short buffer[SAMPLE_LENGTH];	/* one tenth of a second worth of short ints */
		int rc, loop;
		long unsigned int tot_level = 0; /* for calculating average soundlevel */
		int max_level = 0, min_level = 65535;
		int n_threshold = 0;	/* number of samples with level above threshold */

		/* print something which indicates the current recording level */
		if (prevlevel != level)
		{
			if (prevlevel != -1)
				mvwprintw(win, 8, 2 + prevlevel, " ");

			prevlevel = level;
		}
		mvwprintw(win, 8, 2 + level, "|");
		wnoutrefresh(win);
		doupdate();

		/* init structure which makes select() wait for one tenth of a second */
		tv.tv_sec = 0;
		tv.tv_usec = 100000;
		/* init structure which makes select() wait for keypresses */
		FD_ZERO(&rfds);
		FD_SET(0, &rfds);

		/* wait */
		if (select(1, &rfds, NULL, NULL, &tv) == -1)
			error_exit("select()-call failed\n");

		/* key pressed? */
		if (FD_ISSET(0, &rfds))
		{
			c = getch();

			if (c == KEY_LEFT && level > 0)
			{
				level--;
			}
			else if (c == KEY_RIGHT && level < (76-1))
			{
				level++;
			}
			else if (c == KEY_ENTER || c == 13 || toupper(c) == 'Q' || toupper(c) == 'X')
			{
				break;
			}
			else
			{
				flash();
			}

			detect_level = (level * MAX_CHECK_LEVEL) / 76;
			mvwprintw(win, 6, 25, "%d     ", detect_level);
			continue;
		}

		/* check audio-level */
		/* read data */
		rc = READ(fd, (char *)buffer, sizeof(buffer));
		if (rc < 0)
			error_exit("error while reading from audio-device!\n", rc);

		/* check data */
		for(loop=0; loop<SAMPLE_LENGTH; loop++)
		{
			int sample = abs(buffer[loop]);

			if (sample > detect_level)
				n_threshold++;

			tot_level += sample;
			min_level = min(min_level, sample);
			max_level = max(max_level, sample);
		}

		if (n_threshold)
		{
			wattron(win, A_REVERSE);
			mvwprintw(win, 7, 2, "-SOUND!-");
			wattroff(win, A_REVERSE);
			waddch(win, ' ');
		}
		else
		{
			mvwprintw(win, 7, 2, "<silence>");
		}

		mvwprintw(win, 9, 25, "%d     ", tot_level / SAMPLE_LENGTH);
		mvwprintw(win, 10, 2, "Level min: %d, max: %d          ", min_level, max_level);
		mvwprintw(win, 11, 22, "%d     ", max(1, n_threshold));

		/* redraw screen */
		wnoutrefresh(win);
		doupdate();
	}

	delwin(win);
	endwin();

	close(fd);

	return 0;
}
